from fnmatch import fnmatch

import reactivex
from reactivex.abc import ObserverBase, SchedulerBase

from labels.model.parser import Request
from labels.parsers.cataloger.alpine.parse_apk_db import parse_apk_db
from labels.parsers.cataloger.arch.parse_alpm import parse_alpm_db
from labels.parsers.cataloger.debian.parse_dpkg_db import parse_dpkg_db
from labels.parsers.cataloger.redhat.parse_rpm_db import parse_rpm_db


def on_next_db_file(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if "lib/apk/db/installed" in value:
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_apk_db,
                            parser_name="apk-db-selector",
                        ),
                    )
                elif any(
                    fnmatch(value, x)
                    for x in (
                        "**/var/lib/dpkg/status",
                        "*var/lib/dpkg/status",
                        "/var/lib/dpkg/status",
                        "**/var/lib/dpkg/status.d/*",
                        "*var/lib/dpkg/status.d/*",
                        "/var/lib/dpkg/status.d/*",
                        "**/lib/opkg/info/*.control",
                        "*lib/opkg/info/*.control",
                        "/lib/opkg/info/*.control",
                        "**/lib/opkg/status",
                        "*lib/opkg/status",
                        "/lib/opkg/status",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_dpkg_db,
                            parser_name="dpkg-db-selector",
                        ),
                    )
                elif any(
                    fnmatch(value, pattern)
                    for pattern in (
                        "**/var/lib/pacman/local/**/desc",
                        "var/lib/pacman/local/**/desc",
                        "/var/lib/pacman/local/**/desc",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_alpm_db,
                            parser_name="alpm-db-selector",
                        ),
                    )
                elif any(
                    fnmatch(value, x)
                    for x in (
                        (
                            "**/{var/lib,usr/share,usr/lib/sysimage}"
                            "/rpm/{Packages,Packages.db,rpmdb.sqlite}"
                        ),
                        (
                            "/{var/lib,usr/share,usr/lib/sysimage}"
                            "/rpm/{Packages,Packages.db,rpmdb.sqlite}"
                        ),
                        "**/rpmdb.sqlite",
                        "**/var/lib/rpm/Packages",
                        "**/var/lib/rpm/Packages.db",
                        "**/var/lib/rpm/rpmdb.sqlite",
                        "**/usr/share/rpm/Packages",
                        "**/usr/share/rpm/Packages.db",
                        "**/usr/share/rpm/rpmdb.sqlite",
                        "**/usr/lib/sysimage/rpm/Packages",
                        "**/usr/lib/sysimage/rpm/Packages.db",
                        "**/usr/lib/sysimage/rpm/rpmdb.sqlite",
                        "/var/lib/rpm/Packages",
                        "/var/lib/rpm/Packages.db",
                        "/var/lib/rpm/rpmdb.sqlite",
                        "/usr/share/rpm/Packages",
                        "/usr/share/rpm/Packages.db",
                        "/usr/share/rpm/rpmdb.sqlite",
                        "/usr/lib/sysimage/rpm/Packages",
                        "/usr/lib/sysimage/rpm/Packages.db",
                        "/usr/lib/sysimage/rpm/rpmdb.sqlite",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_rpm_db,
                            parser_name="environment-parser",
                        ),
                    )
            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



from labels.model.file import Coordinates, Location
from labels.model.package import Language
from labels.model.syft_sbom import SyftLocation


def get_language(language_str: str) -> Language:
    try:
        return Language(language_str.lower())
    except ValueError:
        return Language.UNKNOWN_LANGUAGE


def get_locations(locations: list[SyftLocation]) -> list[Location]:
    return [
        Location(
            access_path=location.access_path,
            coordinates=Coordinates(
                real_path=location.access_path,
                file_system_id=location.layer_id,
            ),
        )
        for location in locations
        if (annotation := location.annotations)
        and (evidence := annotation.evidence)
        and evidence == "primary"
    ]



import logging
from datetime import datetime
from typing import NotRequired, TypedDict, cast
from xml.etree import ElementTree as ET

from defusedxml import ElementTree
from pydantic import BaseModel, ConfigDict

from labels.config.cache import dual_cache
from labels.enrichers.api_interface import make_get

LOGGER = logging.getLogger(__name__)


class MavenSearchRawDoc(TypedDict):
    id: str
    g: str
    a: str
    timestamp: int
    ec: list[str]
    v: NotRequired[str]
    latestVersion: NotRequired[str]
    p: NotRequired[str]
    tags: NotRequired[list[str]]


class MavenSearchDocResponse(BaseModel):
    id_: str
    group: str
    artifact: str
    version: str
    timestamp: int
    extra_classifiers: list[str]
    packaging: str | None = None
    tags: list[str] | None = None
    model_config = ConfigDict(frozen=True)


class MavenPackageInfo(BaseModel):
    group_id: str
    artifact_id: str
    latest_version: str | None = None
    release_date: int | None = None
    authors: list[str] | None = None
    version: str | None = None
    jar_url: str | None = None
    hash: str | None = None
    model_config = ConfigDict(frozen=True)


def build_maven_search_url(
    artifact_id: str,
    version: str | None,
) -> tuple[str, dict[str, str | int]]:
    base_url = "https://central.sonatype.com/solrsearch/select"
    query = f"a:{artifact_id} AND v:{version}" if version else f"a:{artifact_id}"
    params: dict[str, str | int] = {"q": query, "rows": 5, "wt": "json"}
    return base_url, params


def parse_maven_search_response(
    pkg_raw_data: MavenSearchRawDoc,
    artifact_id: str,
    version: str,
) -> MavenSearchDocResponse | None:
    try:
        return MavenSearchDocResponse(
            id_=pkg_raw_data["id"],
            group=pkg_raw_data["g"],
            artifact=pkg_raw_data["a"],
            version=pkg_raw_data["v"] if "v" in pkg_raw_data else pkg_raw_data["latestVersion"],
            packaging=pkg_raw_data.get("p"),
            timestamp=int(pkg_raw_data["timestamp"] // 1000),
            extra_classifiers=pkg_raw_data["ec"],
            tags=pkg_raw_data.get("tags"),
        )
    except KeyError as exc:
        LOGGER.exception(
            "Error parsing Maven search response",
            extra={
                "extra": {
                    "artifact_id": artifact_id,
                    "version": version,
                    "doc": pkg_raw_data,
                    "key": exc.args[0],  # type: ignore[misc]
                },
            },
        )
        return None


@dual_cache
def get_maven_search_response(artifact_id: str, version: str) -> MavenSearchRawDoc | None:
    base_url, params = build_maven_search_url(artifact_id, version)
    package_data = make_get(base_url, params=params, timeout=30)  # type: ignore[misc]
    if not package_data:  # type: ignore[misc]
        return None

    docs = package_data["response"].get("docs", [])  # type: ignore[misc]
    if not docs or len(docs) > 1:  # type: ignore[misc]
        return None
    return cast("MavenSearchRawDoc", docs[0])  # type: ignore[misc]


def search_maven_package(artifact_id: str, version: str) -> MavenSearchDocResponse | None:
    maven_raw_data = get_maven_search_response(artifact_id, version)
    if not maven_raw_data:
        return None
    return parse_maven_search_response(maven_raw_data, artifact_id, version)


def build_maven_urls(group_id: str, artifact_id: str, version: str) -> tuple[str, str, str]:
    group_id_path = group_id.replace(".", "/")
    base_path = f"https://repo1.maven.org/maven2/{group_id_path}/{artifact_id}/{version}"
    pom_url = f"{base_path}/{artifact_id}-{version}.pom"
    jar_url = f"{base_path}/{artifact_id}-{version}.jar"
    hash_url = f"{base_path}/{artifact_id}-{version}.jar.sha1"
    return pom_url, jar_url, hash_url


def get_package_hash(hash_url: str) -> str:
    hash_response: str | None = make_get(hash_url, content=True, timeout=30)
    return hash_response.strip() if hash_response else "Hash not available"


def parse_pom_xml(pom_content: str) -> ET.Element:
    return ElementTree.fromstring(pom_content)


def extract_authors(pom_xml: ET.Element) -> list[str]:
    namespace = {"m": "http://maven.apache.org/POM/4.0.0"}
    return [
        author.text
        for author in pom_xml.findall("m:developers/m:developer/m:name", namespace)
        if author.text
    ]


@dual_cache
def get_authors(group_id: str, artifact_id: str, version: str) -> list[str] | None:
    pom_url, _, _ = build_maven_urls(group_id, artifact_id, version)
    response = make_get(pom_url, content=True, timeout=30)
    if not response:
        return None

    pom_xml = parse_pom_xml(response)
    return extract_authors(pom_xml)


def parse_metadata_xml(metadata_content: str) -> tuple[str | None, int | None]:
    metadata_xml = ElementTree.fromstring(metadata_content)
    latest_stable_version = metadata_xml.find("versioning/release")
    release_date_tag = metadata_xml.find("versioning/lastUpdated")

    release_date: int | None = None
    if release_date_tag is not None and release_date_tag.text:
        release_date = int(
            datetime.strptime(release_date_tag.text + "+0000", "%Y%m%d%H%M%S%z").timestamp(),
        )

    return (
        latest_stable_version.text if latest_stable_version is not None else None,
        release_date,
    )


def build_metadata_url(group_id: str, artifact_id: str) -> str:
    group_id_path = group_id.replace(".", "/")
    return f"https://repo1.maven.org/maven2/{group_id_path}/{artifact_id}/maven-metadata.xml"


@dual_cache
def get_maven_metadata(group_id: str, artifact_id: str) -> tuple[str | None, int | None]:
    metadata_url = build_metadata_url(group_id, artifact_id)
    response = make_get(metadata_url, content=True, timeout=30)
    if not response:
        return None, None

    return parse_metadata_xml(response)


@dual_cache
def get_package_hash_from_pom(
    group_id: str,
    artifact_id: str,
    version: str,
) -> tuple[str | None, str]:
    pom_url, jar_url, hash_url = build_maven_urls(group_id, artifact_id, version)
    response = make_get(pom_url, timeout=30, content=True)
    if not response:
        return None, jar_url

    return get_package_hash(hash_url), jar_url


def get_specific_version_info(
    group_id: str,
    artifact_id: str,
    version: str,
) -> MavenPackageInfo | None:
    package_hash, jar_url = get_package_hash_from_pom(group_id, artifact_id, version)

    if not package_hash:
        return None

    return MavenPackageInfo(
        group_id=group_id,
        artifact_id=artifact_id,
        version=version,
        jar_url=jar_url,
        hash=package_hash,
    )


def get_latest_version_info(group_id: str, artifact_id: str) -> MavenPackageInfo | None:
    latest_version, release_date = get_maven_metadata(group_id, artifact_id)
    if not latest_version:
        return None

    authors = get_authors(group_id, artifact_id, latest_version)

    return MavenPackageInfo(
        group_id=group_id,
        artifact_id=artifact_id,
        latest_version=latest_version,
        release_date=release_date,
        authors=authors if authors else None,
    )


def get_maven_package_info(
    group_id: str,
    artifact_id: str,
    version: str | None = None,
) -> MavenPackageInfo | None:
    if version:
        return get_specific_version_info(group_id, artifact_id, version)
    return get_latest_version_info(group_id, artifact_id)



import threading
from typing import Any, Literal, overload

import requests
from requests import Response
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


class _ThreadLocalSession(threading.local):
    session: requests.Session | None = None


_thread_local = _ThreadLocalSession()


def _get_session_with_retry(
    retries: int = 3,
    backoff_factor: float = 0.5,
    status_forcelist: tuple[int, ...] = (500, 502, 503, 504),
    pool_connections: int = 10,
    pool_maxsize: int = 10,
) -> requests.Session:
    session = requests.Session()
    retry_strategy = Retry(
        total=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
        allowed_methods=["GET"],
    )
    adapter = HTTPAdapter(
        max_retries=retry_strategy,
        pool_connections=pool_connections,
        pool_maxsize=pool_maxsize,
    )
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


def get_session() -> requests.Session:
    if _thread_local.session is None:
        _thread_local.session = _get_session_with_retry()
    return _thread_local.session


@overload
def make_get(  # type: ignore[explicit-any]  # pragma: no cover
    url: str,
    *,
    timeout: float = 30,
    content: Literal[True],
    **kwargs: Any,  # noqa: ANN401
) -> str | None: ...


@overload
def make_get(  # type: ignore[explicit-any]  # pragma: no cover
    url: str,
    *,
    timeout: float = 30,
    content: Literal[False] = False,
    **kwargs: Any,  # noqa: ANN401
) -> Any | None: ...  # noqa: ANN401


def make_get(  # type: ignore[explicit-any]
    url: str,
    *,
    timeout: float = 30,
    content: bool = False,
    **kwargs: Any,
) -> Any | None:
    response: Response = get_session().get(url, timeout=timeout, **kwargs)  # type: ignore[misc]

    try:
        if response.status_code != 200:
            return None

        if content:
            return response.content.decode("utf-8")

        return response.json()  # type: ignore[misc]
    except ValueError:
        return None
    finally:
        response.close()



from datetime import datetime

from pydantic import BaseModel

from labels.model.constraints import NonEmptyStr


class Digest(BaseModel):
    algorithm: NonEmptyStr | None = None
    value: NonEmptyStr | None = None


class Artifact(BaseModel):
    url: NonEmptyStr
    integrity: Digest | None = None


class HealthMetadata(BaseModel):
    latest_version: NonEmptyStr | None = None
    latest_version_created_at: NonEmptyStr | datetime | None = None
    artifact: Artifact | None = None
    authors: NonEmptyStr | None = None
